"""Module with data plugins that represent files of completed calculations jobs that have been stashed."""

# AUTO-GENERATED

# fmt: off

from .base import *
from .folder import *

__all__ = (
    'RemoteStashData',
    'RemoteStashFolderData',
)

# fmt: on



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Data class that can be used to store a single file in its repository."""

from __future__ import annotations

import contextlib
import io
import os
import pathlib
import typing as t

from aiida.common import exceptions

from .data import Data

__all__ = ('SinglefileData',)

FilePath = t.Union[str, pathlib.PurePosixPath]


class SinglefileData(Data):
    """Data class that can be used to store a single file in its repository."""

    DEFAULT_FILENAME = 'file.txt'

    @classmethod
    def from_string(cls, content: str, filename: str | pathlib.Path | None = None, **kwargs: t.Any) -> 'SinglefileData':
        """Construct a new instance and set ``content`` as its contents.

        :param content: The content as a string.
        :param filename: Specify filename to use (defaults to ``file.txt``).
        """
        return cls(io.StringIO(content), filename, **kwargs)

    def __init__(
        self, file: str | pathlib.Path | t.IO, filename: str | pathlib.Path | None = None, **kwargs: t.Any
    ) -> None:
        """Construct a new instance and set the contents to that of the file.

        :param file: an absolute filepath or filelike object whose contents to copy.
            Hint: Pass io.BytesIO(b"my string") to construct the SinglefileData directly from a string.
        :param filename: specify filename to use (defaults to name of provided file).
        """
        super().__init__(**kwargs)

        if file is not None:
            self.set_file(file, filename=filename)

    @property
    def filename(self) -> str:
        """Return the name of the file stored.

        :return: the filename under which the file is stored in the repository
        """
        return self.base.attributes.get('filename')

    @t.overload
    @contextlib.contextmanager
    def open(self, path: FilePath, mode: t.Literal['r'] = ...) -> t.Iterator[t.TextIO]: ...

    @t.overload
    @contextlib.contextmanager
    def open(self, path: FilePath, mode: t.Literal['rb']) -> t.Iterator[t.BinaryIO]: ...

    @t.overload
    @contextlib.contextmanager
    def open(  # type: ignore[overload-overlap]
        self, path: None = None, mode: t.Literal['r'] = ...
    ) -> t.Iterator[t.TextIO]: ...

    @t.overload
    @contextlib.contextmanager
    def open(self, path: None = None, mode: t.Literal['rb'] = ...) -> t.Iterator[t.BinaryIO]: ...

    @contextlib.contextmanager
    def open(
        self, path: FilePath | None = None, mode: t.Literal['r', 'rb'] = 'r'
    ) -> t.Iterator[t.BinaryIO] | t.Iterator[t.TextIO]:
        """Return an open file handle to the content of this data node.

        :param path: the relative path of the object within the repository.
        :param mode: the mode with which to open the file handle (default: read mode)
        :return: a file handle
        """
        if path is None:
            path = self.filename

        with self.base.repository.open(path, mode=mode) as handle:
            yield handle

    @contextlib.contextmanager
    def as_path(self) -> t.Iterator[pathlib.Path]:
        """Make the contents of the file available as a normal filepath on the local file system.

        :param path: optional relative path of the object within the repository.
        :return: the filepath of the content of the repository or object if ``path`` is specified.
        :raises TypeError: if the path is not a string or ``Path``, or is an absolute path.
        :raises FileNotFoundError: if no object exists for the given path.
        """
        with self.base.repository.as_path(self.filename) as filepath:
            yield filepath

    def get_content(self, mode: str = 'r') -> str | bytes:
        """Return the content of the single file stored for this data node.

        :param mode: the mode with which to open the file handle (default: read mode)
        :return: the content of the file as a string or bytes, depending on ``mode``.
        """
        with self.open(mode=mode) as handle:  # type: ignore[call-overload]
            return handle.read()

    def set_file(self, file: str | pathlib.Path | t.IO, filename: str | pathlib.Path | None = None) -> None:
        """Store the content of the file in the node's repository, deleting any other existing objects.

        :param file: an absolute filepath or filelike object whose contents to copy
            Hint: Pass io.BytesIO(b"my string") to construct the file directly from a string.
        :param filename: specify filename to use (defaults to name of provided file).
        """
        if isinstance(file, (str, pathlib.Path)):
            is_filelike = False

            key = os.path.basename(file)
            if not os.path.isabs(file):
                raise ValueError(f'path `{file}` is not absolute')

            if not os.path.isfile(file):
                raise ValueError(f'path `{file}` does not correspond to an existing file')
        else:
            is_filelike = True
            try:
                key = os.path.basename(file.name)
            except AttributeError:
                key = self.DEFAULT_FILENAME

        key = str(filename) if filename is not None else key
        existing_object_names = self.base.repository.list_object_names()

        try:
            # Remove the 'key' from the list of currently existing objects such that it is not deleted after storing
            existing_object_names.remove(key)
        except ValueError:
            pass

        if is_filelike:
            self.base.repository.put_object_from_filelike(file, key)  # type: ignore[arg-type]
        else:
            self.base.repository.put_object_from_file(file, key)  # type: ignore[arg-type]

        # Delete any other existing objects (minus the current `key` which was already removed from the list)
        for existing_key in existing_object_names:
            self.base.repository.delete_object(existing_key)

        self.base.attributes.set('filename', key)

    def _validate(self) -> bool:
        """Ensure that there is one object stored in the repository, whose key matches value set for `filename` attr."""
        super()._validate()

        try:
            filename = self.filename
        except AttributeError:
            raise exceptions.ValidationError('the `filename` attribute is not set.')

        objects = self.base.repository.list_object_names()

        if [filename] != objects:
            raise exceptions.ValidationError(
                f'respository files {objects} do not match the `filename` attribute `{filename}`.'
            )

        return True



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Module to manage authentification information for the SQLA backend."""

from sqlalchemy import ForeignKey
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.orm import backref, relationship
from sqlalchemy.schema import Column, UniqueConstraint
from sqlalchemy.types import Boolean, Integer

from .base import Base


class DbAuthInfo(Base):
    """Database model to store data for :py:class:`aiida.orm.AuthInfo`, and keep computer authentication data, per user.

    Specifications are user-specific of how to submit jobs in the computer.
    The model also has an ``enabled`` logical switch that indicates whether the device is available for use or not.
    This last one can be set and unset by the user.
    """

    __tablename__ = 'db_dbauthinfo'

    id = Column(Integer, primary_key=True)
    aiidauser_id = Column(
        Integer,
        ForeignKey('db_dbuser.id', ondelete='CASCADE', deferrable=True, initially='DEFERRED'),
        nullable=False,
        index=True,
    )
    dbcomputer_id = Column(
        Integer,
        ForeignKey('db_dbcomputer.id', ondelete='CASCADE', deferrable=True, initially='DEFERRED'),
        nullable=False,
        index=True,
    )
    _metadata = Column('metadata', JSONB, default=dict, nullable=False)
    auth_params = Column(JSONB, default=dict, nullable=False)
    enabled = Column(Boolean, default=True, nullable=False)

    aiidauser = relationship('DbUser', backref=backref('authinfos', passive_deletes=True, cascade='all, delete'))
    dbcomputer = relationship('DbComputer', backref=backref('authinfos', passive_deletes=True, cascade='all, delete'))

    __table_args__ = (UniqueConstraint('aiidauser_id', 'dbcomputer_id'),)

    def __str__(self):
        if self.enabled:
            return f'DB authorization info for {self.aiidauser.email} on {self.dbcomputer.label}'
        return f'DB authorization info for {self.aiidauser.email} on {self.dbcomputer.label} [DISABLED]'



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Common variables"""

import os
import shutil
import tempfile
from pathlib import Path
from typing import Callable, Sequence

from archive_path import TarPath, ZipPath

from aiida.common import exceptions
from aiida.common.progress_reporter import create_callback, get_progress_reporter


def update_metadata(metadata, version):
    """Update the metadata with a new version number and a notification of the conversion that was executed.

    :param metadata: the content of an export archive metadata.json file
    :param version: string version number that the updated metadata should get
    """
    from aiida import get_version

    old_version = metadata['export_version']
    conversion_info = metadata.get('conversion_info', [])

    conversion_message = f'Converted from version {old_version} to {version} with AiiDA v{get_version()}'
    conversion_info.append(conversion_message)

    metadata['aiida_version'] = get_version()
    metadata['export_version'] = version
    metadata['conversion_info'] = conversion_info


def verify_metadata_version(metadata, version=None):
    """Utility function to verify that the metadata has the correct version number.

    If no version number is passed, it will just extract the version number and return it.

    :param metadata: the content of an export archive metadata.json file
    :param version: string version number that the metadata is expected to have
    """
    try:
        metadata_version = metadata['export_version']
    except KeyError:
        raise exceptions.StorageMigrationError("metadata is missing the 'export_version' key")

    if version is None:
        return metadata_version

    if metadata_version != version:
        raise exceptions.StorageMigrationError(
            f'expected archive file with version {version} but found version {metadata_version}'
        )

    return None


def copy_zip_to_zip(
    inpath: Path,
    outpath: Path,
    path_callback: Callable[[ZipPath, ZipPath], bool],
    *,
    compression: int = 6,
    overwrite: bool = True,
    title: str = 'Writing new zip file',
    info_order: Sequence[str] = (),
) -> None:
    """Create a new zip file from an existing zip file.

    All files/folders are streamed directly to the new zip file,
    with the ``path_callback`` allowing for per path modifications.
    The new zip file is first created in a temporary directory, and then moved to the desired location.

    :param inpath: the path to the existing archive
    :param outpath: the path to output the new archive
    :param path_callback: a callback that is called for each path in the archive: ``(inpath, outpath) -> handled``
        If handled is ``True``, the path is assumed to already have been copied to the new zip file.
    :param compression: the default compression level to use for the new zip file
    :param overwrite: whether to overwrite the output file if it already exists
    :param title: the title of the progress bar
    :param info_order: ``ZipInfo`` for these file names will be written first to the zip central directory.
        This allows for faster reading of these files, with ``archive_path.read_file_in_zip``.
    """
    if (not overwrite) and outpath.exists() and outpath.is_file():
        raise FileExistsError(f'{outpath} already exists')
    with tempfile.TemporaryDirectory() as tmpdirname:
        temp_archive = Path(tmpdirname) / 'archive.zip'
        with ZipPath(temp_archive, mode='w', compresslevel=compression, info_order=info_order) as new_path:
            with ZipPath(inpath, mode='r') as path:
                length = sum(1 for _ in path.glob('**/*', include_virtual=False))
                with get_progress_reporter()(desc=title, total=length) as progress:
                    for subpath in path.glob('**/*', include_virtual=False):
                        new_path_sub = new_path.joinpath(subpath.at)
                        if path_callback(subpath, new_path_sub):
                            pass
                        elif subpath.is_dir():
                            new_path_sub.mkdir(exist_ok=True)
                        else:
                            new_path_sub.putfile(subpath)
                        progress.update()
        if overwrite and outpath.exists() and outpath.is_file():
            outpath.unlink()
        shutil.move(temp_archive, outpath)


def copy_tar_to_zip(
    inpath: Path,
    outpath: Path,
    path_callback: Callable[[Path, ZipPath], bool],
    *,
    compression: int = 6,
    overwrite: bool = True,
    title: str = 'Writing new zip file',
    info_order: Sequence[str] = (),
) -> None:
    """Create a new zip file from an existing tar file.

    The tar file is first extracted to a temporary directory, and then the new zip file is created,
    with the ``path_callback`` allowing for per path modifications.
    The new zip file is first created in a temporary directory, and then moved to the desired location.

    :param inpath: the path to the existing archive
    :param outpath: the path to output the new archive
    :param path_callback: a callback that is called for each path in the archive: ``(inpath, outpath) -> handled``
        If handled is ``True``, the path is assumed to already have been copied to the new zip file.
    :param compression: the default compression level to use for the new zip file
    :param overwrite: whether to overwrite the output file if it already exists
    :param title: the title of the progress bar
    :param info_order: ``ZipInfo`` for these file names will be written first to the zip central directory.
        This allows for faster reading of these files, with ``archive_path.read_file_in_zip``.
    """
    if (not overwrite) and outpath.exists() and outpath.is_file():
        raise FileExistsError(f'{outpath} already exists')
    with tempfile.TemporaryDirectory() as tmpdirname:
        # for tar files we extract first, since the file is compressed as a single object
        temp_extracted = Path(tmpdirname) / 'extracted'
        with get_progress_reporter()(total=1) as progress:
            callback = create_callback(progress)
            with TarPath(inpath, mode='r:*') as path:
                path.extract_tree(
                    temp_extracted,
                    allow_dev=False,
                    allow_symlink=False,
                    callback=callback,
                    cb_descript=f'{title} (extracting tar)',
                )
        temp_archive = Path(tmpdirname) / 'archive.zip'
        with ZipPath(temp_archive, mode='w', compresslevel=compression, info_order=info_order) as new_path:
            length = sum(1 for _ in temp_extracted.glob('**/*'))
            with get_progress_reporter()(desc=title, total=length) as progress:
                for subpath in temp_extracted.glob('**/*'):
                    new_path_sub = new_path.joinpath(subpath.relative_to(temp_extracted).as_posix())
                    if path_callback(subpath.relative_to(temp_extracted), new_path_sub):
                        pass
                    elif subpath.is_dir():
                        new_path_sub.mkdir(exist_ok=True)
                    else:
                        # files extracted from the tar do not include a modified time, yet zip requires one
                        os.utime(subpath, (subpath.stat().st_ctime, subpath.stat().st_ctime))
                        new_path_sub.putfile(subpath)
                    progress.update()

        if overwrite and outpath.exists() and outpath.is_file():
            outpath.unlink()
        shutil.move(temp_archive, outpath)



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Functionality for dumping of ProcessNodes."""

from __future__ import annotations

import logging
from pathlib import Path
from types import SimpleNamespace
from typing import List

import yaml

from aiida.common import LinkType
from aiida.common.exceptions import NotExistentAttributeError
from aiida.orm import (
    CalcFunctionNode,
    CalcJobNode,
    CalculationNode,
    LinkManager,
    ProcessNode,
    WorkChainNode,
    WorkflowNode,
    WorkFunctionNode,
)
from aiida.orm.utils import LinkTriple

LOGGER = logging.getLogger(__name__)


class ProcessDumper:
    def __init__(
        self,
        include_inputs: bool = True,
        include_outputs: bool = False,
        include_attributes: bool = True,
        include_extras: bool = True,
        overwrite: bool = False,
        flat: bool = False,
    ) -> None:
        self.include_inputs = include_inputs
        self.include_outputs = include_outputs
        self.include_attributes = include_attributes
        self.include_extras = include_extras
        self.overwrite = overwrite
        self.flat = flat

    @staticmethod
    def _generate_default_dump_path(process_node: ProcessNode) -> Path:
        """Simple helper function to generate the default parent-dumping directory if none given.

        This function is not called for the recursive sub-calls of `_dump_calculation` as it just creates the default
        parent folder for the dumping, if no name is given.

        :param process_node: The `ProcessNode` for which the directory is created.
        :return: The absolute default parent dump path.
        """

        pk = process_node.pk
        try:
            return Path(f'dump-{process_node.process_label}-{pk}')
        except AttributeError:
            # This case came up during testing, not sure how relevant it actually is
            return Path(f'dump-{process_node.process_type}-{pk}')

    @staticmethod
    def _generate_readme(process_node: ProcessNode, output_path: Path) -> None:
        """Generate README.md file in main dumping directory.

        :param process_node: `CalculationNode` or `WorkflowNode`.
        :param output_path: Output path for dumping.

        """

        import textwrap

        from aiida.cmdline.utils.ascii_vis import format_call_graph
        from aiida.cmdline.utils.common import (
            get_calcjob_report,
            get_node_info,
            get_process_function_report,
            get_workchain_report,
        )

        pk = process_node.pk

        _readme_string = textwrap.dedent(
            f"""\
        This directory contains the files involved in the calculation/workflow
        `{process_node.process_label} <{pk}>` run with AiiDA.

        Child calculations/workflows (also called `CalcJob`s/`CalcFunction`s and `WorkChain`s/`WorkFunction`s in AiiDA
        jargon) run by the parent workflow are contained in the directory tree as sub-folders and are sorted by their
        creation time. The directory tree thus mirrors the logical execution of the workflow, which can also be queried
        by running `verdi process status {pk}` on the command line.

        By default, input and output files of each calculation can be found in the corresponding "inputs" and "outputs"
        directories (the former also contains the hidden ".aiida" folder with machine-readable job execution settings).
        Additional input and output files (depending on the type of calculation) are placed in the "node_inputs" and
        "node_outputs", respectively.

        Lastly, every folder also contains a hidden, human-readable `.aiida_node_metadata.yaml` file with the relevant
        AiiDA node data for further inspection."""
        )

        # `verdi process status`
        process_status = format_call_graph(calc_node=process_node, max_depth=None, call_link_label=True)
        _readme_string += f'\n\n\nOutput of `verdi process status {pk}`:\n\n```shell\n{process_status}\n```'

        # `verdi process report`
        # Copied over from `cmd_process`
        if isinstance(process_node, CalcJobNode):
            process_report = get_calcjob_report(process_node)
        elif isinstance(process_node, WorkChainNode):
            process_report = get_workchain_report(process_node, levelname='REPORT', indent_size=2, max_depth=None)
        elif isinstance(process_node, (CalcFunctionNode, WorkFunctionNode)):
            process_report = get_process_function_report(process_node)
        else:
            process_report = f'Nothing to show for node type {process_node.__class__}'

        _readme_string += f'\n\n\nOutput of `verdi process report {pk}`:\n\n```shell\n{process_report}\n```'

        # `verdi process show`?
        process_show = get_node_info(node=process_node)
        _readme_string += f'\n\n\nOutput of `verdi process show {pk}`:\n\n```shell\n{process_show}\n```'

        (output_path / 'README.md').write_text(_readme_string)

    @staticmethod
    def _generate_child_node_label(index: int, link_triple: LinkTriple) -> str:
        """Small helper function to generate and clean directory label for child nodes during recursion.

        :param index: Index assigned to step at current level of recursion.
        :param link_triple: `LinkTriple` of `ProcessNode` explored during recursion.
        :return: Chlild node label during recursion.
        """
        node = link_triple.node
        link_label = link_triple.link_label

        # Generate directories with naming scheme akin to `verdi process status`
        label_list = [f'{index:02d}', link_label]

        try:
            process_label = node.process_label
            if process_label is not None and process_label != link_label:
                label_list += [process_label]

        except AttributeError:
            process_type = node.process_type
            if process_type is not None and process_type != link_label:
                label_list += [process_type]

        node_label = '-'.join(label_list)
        # `CALL-` as part of the link labels also for MultiplyAddWorkChain -> Seems general enough, so remove
        node_label = node_label.replace('CALL-', '')
        node_label = node_label.replace('None-', '')

        return node_label

    def dump(
        self,
        process_node: ProcessNode,
        output_path: Path | None,
        io_dump_paths: List[str | Path] | None = None,
    ) -> Path:
        """Dumps all data involved in a `ProcessNode`, including its outgoing links.

        Note that if an outgoing link is a `WorkflowNode`, the function recursively calls itself, while files are
        only actually created when a `CalculationNode` is reached.

        :param process_node: The parent `ProcessNode` node to be dumped.
        :param output_path: The output path where the directory tree will be created.
        :param io_dump_paths: Subdirectories created for each `CalculationNode`.
            Default: ['inputs', 'outputs', 'node_inputs', 'node_outputs']
        """

        if output_path is None:
            output_path = self._generate_default_dump_path(process_node=process_node)

        self._validate_make_dump_path(validate_path=output_path)

        if isinstance(process_node, CalculationNode):
            self._dump_calculation(
                calculation_node=process_node,
                output_path=output_path,
                io_dump_paths=io_dump_paths,
            )

        elif isinstance(process_node, WorkflowNode):
            self._dump_workflow(
                workflow_node=process_node,
                output_path=output_path,
                io_dump_paths=io_dump_paths,
            )

        self._generate_readme(process_node=process_node, output_path=output_path)

        return output_path

    def _dump_workflow(
        self, workflow_node: WorkflowNode, output_path: Path, io_dump_paths: List[str | Path] | None = None
    ) -> None:
        """Recursive function to traverse a `WorkflowNode` and dump its `CalculationNode` s.

        :param workflow_node: `WorkflowNode` to be traversed. Will be updated during recursion.
        :param output_path: Dumping parent directory. Will be updated during recursion.
        :param io_dump_paths: Custom subdirectories for `CalculationNode` s, defaults to None
        """

        self._validate_make_dump_path(validate_path=output_path)
        self._dump_node_yaml(process_node=workflow_node, output_path=output_path)

        called_links = workflow_node.base.links.get_outgoing(link_type=(LinkType.CALL_CALC, LinkType.CALL_WORK)).all()
        called_links = sorted(called_links, key=lambda link_triple: link_triple.node.ctime)

        for index, link_triple in enumerate(called_links, start=1):
            child_node = link_triple.node
            child_label = self._generate_child_node_label(index=index, link_triple=link_triple)
            child_output_path = output_path.resolve() / child_label

            # Recursive function call for `WorkFlowNode`
            if isinstance(child_node, WorkflowNode):
                self._dump_workflow(
                    workflow_node=child_node,
                    output_path=child_output_path,
                    io_dump_paths=io_dump_paths,
                )

            # Once a `CalculationNode` as child reached, dump it
            elif isinstance(child_node, CalculationNode):
                self._dump_calculation(
                    calculation_node=child_node,
                    output_path=child_output_path,
                    io_dump_paths=io_dump_paths,
                )

    def _dump_calculation(
        self,
        calculation_node: CalculationNode,
        output_path: Path,
        io_dump_paths: List[str | Path] | None = None,
    ) -> None:
        """Dump the contents of a `CalculationNode` to a specified output path.

        :param calculation_node: The `CalculationNode` to be dumped.
        :param output_path: The path where the files will be dumped.
        :param io_dump_paths: Subdirectories created for the `CalculationNode`.
            Default: ['inputs', 'outputs', 'node_inputs', 'node_outputs']
        """

        self._validate_make_dump_path(validate_path=output_path)
        self._dump_node_yaml(process_node=calculation_node, output_path=output_path)

        io_dump_mapping = self._generate_calculation_io_mapping(io_dump_paths=io_dump_paths)

        # Dump the repository contents of the node
        calculation_node.base.repository.copy_tree(output_path.resolve() / io_dump_mapping.repository)

        # Dump the repository contents of `outputs.retrieved`
        try:
            calculation_node.outputs.retrieved.base.repository.copy_tree(
                output_path.resolve() / io_dump_mapping.retrieved
            )
        except NotExistentAttributeError:
            pass

        # Dump the node_inputs
        if self.include_inputs:
            input_links = calculation_node.base.links.get_incoming(link_type=LinkType.INPUT_CALC)
            self._dump_calculation_io(parent_path=output_path / io_dump_mapping.inputs, link_triples=input_links)

        # Dump the node_outputs apart from `retrieved`
        if self.include_outputs:
            output_links = list(calculation_node.base.links.get_outgoing(link_type=LinkType.CREATE))
            output_links = [output_link for output_link in output_links if output_link.link_label != 'retrieved']

            self._dump_calculation_io(
                parent_path=output_path / io_dump_mapping.outputs,
                link_triples=output_links,
            )

    def _dump_calculation_io(self, parent_path: Path, link_triples: LinkManager | List[LinkTriple]):
        """Small helper function to dump linked input/output nodes of a `CalculationNode`.

        :param parent_path: Parent directory for dumping the linked node contents.
        :param link_triples: List of link triples.
        """

        for link_triple in link_triples:
            link_label = link_triple.link_label

            if not self.flat:
                linked_node_path = parent_path / Path(*link_label.split('__'))
            else:
                # Don't use link_label at all -> But, relative path inside FolderData is retained
                linked_node_path = parent_path

            link_triple.node.base.repository.copy_tree(linked_node_path.resolve())

    def _validate_make_dump_path(self, validate_path: Path, safeguard_file: str = '.aiida_node_metadata.yaml') -> Path:
        """Create default dumping directory for a given process node and return it as absolute path.

        :param validate_path: Path to validate for dumping.
        :param safeguard_file: Dumping-specific file to avoid deleting wrong directory.
            Default: `.aiida_node_metadata.yaml`
        :return: The absolute created dump path.
        """
        import shutil

        if validate_path.is_dir():
            # Existing, empty directory -> OK
            if not any(validate_path.iterdir()):
                pass

            # Existing, non-empty directory and overwrite False -> FileExistsError
            elif not self.overwrite:
                raise FileExistsError(f'Path `{validate_path}` already exists and overwrite set to False.')

            # Existing, non-empty directory and overwrite True
            # Check for safeguard file ('.aiida_node_metadata.yaml') for safety
            # If present -> Remove directory
            elif (validate_path / safeguard_file).is_file():
                LOGGER.info(f'Overwrite set to true, will overwrite directory `{validate_path}`.')
                shutil.rmtree(validate_path)

            # Existing and non-empty directory and overwrite True
            # Check for safeguard file ('.aiida_node_metadata.yaml') for safety
            # If absent -> Don't remove directory as to not accidentally remove a wrong one
            else:
                raise Exception(
                    f"Path `{validate_path}` already exists and doesn't contain safeguard file {safeguard_file}."
                    f' Not removing for safety reasons.'
                )

        # Not included in if-else as to avoid having to repeat the `mkdir` call.
        # `exist_ok=True` as checks implemented above
        validate_path.mkdir(exist_ok=True, parents=True)

        return validate_path.resolve()

    def _generate_calculation_io_mapping(self, io_dump_paths: List[str | Path] | None = None) -> SimpleNamespace:
        """Helper function to generate mapping for entities dumped for each `CalculationNode`.

        This is to avoid exposing AiiDA terminology, like `repository` to the user, while keeping track of which
        entities should be dumped into which directory, and allowing for alternative directory names.

        :param io_dump_paths: Subdirectories created for the `CalculationNode`.
            Default: ['inputs', 'outputs', 'node_inputs', 'node_outputs']
        :return: SimpleNamespace mapping.
        """

        aiida_entities_to_dump = ['repository', 'retrieved', 'inputs', 'outputs']
        default_calculation_io_dump_paths = ['inputs', 'outputs', 'node_inputs', 'node_outputs']
        empty_calculation_io_dump_paths = [''] * 4

        if self.flat and io_dump_paths is None:
            LOGGER.info(
                'Flat set to True and no `io_dump_paths`. Dumping in a flat directory, files might be overwritten.'
            )
            return SimpleNamespace(**dict(zip(aiida_entities_to_dump, empty_calculation_io_dump_paths)))

        elif not self.flat and io_dump_paths is None:
            LOGGER.info(
                'Flat set to False but no `io_dump_paths` provided. '
                + f'Will use the defaults {default_calculation_io_dump_paths}.'
            )
            return SimpleNamespace(**dict(zip(aiida_entities_to_dump, default_calculation_io_dump_paths)))

        elif self.flat and io_dump_paths is not None:
            LOGGER.info('Flat set to True but `io_dump_paths` provided. These will be used, but `inputs` not nested.')
            return SimpleNamespace(**dict(zip(aiida_entities_to_dump, io_dump_paths)))
        else:
            LOGGER.info(
                'Flat set to False but no `io_dump_paths` provided. These will be used, but `node_inputs` flattened.'
            )
            return SimpleNamespace(**dict(zip(aiida_entities_to_dump, io_dump_paths)))  # type: ignore[arg-type]

    def _dump_node_yaml(
        self,
        process_node: ProcessNode,
        output_path: Path,
        output_filename: str = '.aiida_node_metadata.yaml',
    ) -> None:
        """Dump the selected `ProcessNode` properties, attributes, and extras to a YAML file.

        :param process_node: The `ProcessNode` to dump.
        :param output_path: The path to the directory where the YAML file will be saved.
        :param output_filename: The name of the output YAML file. Defaults to `.aiida_node_metadata.yaml`.
        """

        node_properties = [
            'label',
            'description',
            'pk',
            'uuid',
            'ctime',
            'mtime',
            'node_type',
            'process_type',
            'is_finished_ok',
        ]

        user_properties = ('first_name', 'last_name', 'email', 'institution')

        computer_properties = ('label', 'hostname', 'scheduler_type', 'transport_type')

        node_dict = {}
        metadata_dict = {}

        # Add actual node `@property`s to dictionary
        for metadata_property in node_properties:
            metadata_dict[metadata_property] = getattr(process_node, metadata_property)

        node_dict['Node data'] = metadata_dict

        # Add user data
        try:
            node_dbuser = process_node.user
            user_dict = {}
            for user_property in user_properties:
                user_dict[user_property] = getattr(node_dbuser, user_property)
            node_dict['User data'] = user_dict
        except AttributeError:
            pass

        # Add computer data
        try:
            node_dbcomputer = process_node.computer
            computer_dict = {}
            for computer_property in computer_properties:
                computer_dict[computer_property] = getattr(node_dbcomputer, computer_property)
            node_dict['Computer data'] = computer_dict
        except AttributeError:
            pass

        # Add node attributes
        if self.include_attributes:
            node_attributes = process_node.base.attributes.all
            node_dict['Node attributes'] = node_attributes

        # Add node extras
        if self.include_extras:
            node_extras = process_node.base.extras.all
            if node_extras:
                node_dict['Node extras'] = node_extras

        output_file = output_path.resolve() / output_filename
        with open(output_file, 'w') as handle:
            yaml.dump(node_dict, handle, sort_keys=False)



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Default values and lazy default get methods for command line options."""

from aiida.cmdline.utils import echo
from aiida.common import exceptions
from aiida.manage.configuration import get_config


def get_default_profile():
    """Try to get the name of the default profile.

    This utility function should only be used for defaults or callbacks in command line interface parameters.
    Otherwise, the preference should go to calling `get_config` to load the actual config and using
    `config.default_profile_name` to get the default profile name.

    :raises click.UsageError: if the config could not be loaded or no default profile exists
    :return: the default profile name or None if no default is defined in the configuration
    """
    try:
        config = get_config(create=True)
    except exceptions.ConfigurationError as exception:
        echo.echo_critical(str(exception))

    try:
        default_profile = config.get_profile(config.default_profile_name).name
    except exceptions.ProfileConfigurationError:
        default_profile = None

    return default_profile



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Module with `Node` sub classes for processes."""

# AUTO-GENERATED

# fmt: off

from .calculation import *
from .process import *
from .workflow import *

__all__ = (
    'CalcFunctionNode',
    'CalcJobNode',
    'CalculationNode',
    'ProcessNode',
    'WorkChainNode',
    'WorkFunctionNode',
    'WorkflowNode',
)

# fmt: on



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Utilities for operations on files on remote computers."""

import os

from aiida.orm.nodes.data.remote.base import RemoteData


def clean_remote(transport, path):
    """Recursively remove a remote folder, with the given absolute path, and all its contents. The path should be
    made accessible through the transport channel, which should already be open

    :param transport: an open Transport channel
    :param path: an absolute path on the remote made available through the transport
    """
    if not isinstance(path, str):
        raise ValueError('the path has to be a string type')

    if not os.path.isabs(path):
        raise ValueError('the path should be absolute')

    if not transport.is_open:
        raise ValueError('the transport should already be open')

    basedir, relative_path = os.path.split(path)

    try:
        transport.chdir(basedir)
        transport.rmtree(relative_path)
    except OSError:
        pass


def get_calcjob_remote_paths(
    pks=None,
    past_days=None,
    older_than=None,
    computers=None,
    user=None,
    backend=None,
    exit_status=None,
    only_not_cleaned=False,
):
    """Return a mapping of computer uuids to a list of remote paths, for a given set of calcjobs. The set of
    calcjobs will be determined by a query with filters based on the pks, past_days, older_than,
    computers and user arguments.

    :param pks: only include calcjobs with a pk in this list
    :param past_days: only include calcjobs created since past_days
    :param older_than: only include calcjobs older than
    :param computers: only include calcjobs that were ran on these computers
    :param user: only include calcjobs of this user
    :param exit_status: only select calcjob with this exit_status
    :param only_not_cleaned: only include calcjobs whose workdir have not been cleaned
    :return: mapping of computer uuid and list of remote folder
    """
    from datetime import timedelta

    from aiida import orm
    from aiida.common import timezone
    from aiida.orm import CalcJobNode

    filters_calc = {}
    filters_computer = {}
    filters_remote = {}

    if user is None:
        if backend:
            user = orm.User.get_collection(backend).get_default()
        else:
            user = orm.User.collection.get_default()

    if computers is not None:
        filters_computer['id'] = {'in': [computer.pk for computer in computers]}

    if past_days is not None:
        filters_calc['mtime'] = {'>': timezone.now() - timedelta(days=past_days)}

    if older_than is not None:
        older_filter = {'<': timezone.now() - timedelta(days=older_than)}
        # Check if we need to apply the AND condition
        if 'mtime' not in filters_calc:
            filters_calc['mtime'] = older_filter
        else:
            past_filter = filters_calc['mtime']
            filters_calc['mtime'] = {'and': [past_filter, older_filter]}

    if exit_status is not None:
        filters_calc['attributes.exit_status'] = exit_status

    if pks:
        filters_calc['id'] = {'in': pks}

    if only_not_cleaned is True:
        filters_remote['or'] = [
            {f'extras.{RemoteData.KEY_EXTRA_CLEANED}': {'!==': True}},
            {'extras': {'!has_key': RemoteData.KEY_EXTRA_CLEANED}},
        ]

    query = orm.QueryBuilder(backend=backend)
    query.append(CalcJobNode, tag='calc', filters=filters_calc)
    query.append(
        RemoteData, tag='remote', project=['*'], edge_filters={'label': 'remote_folder'}, filters=filters_remote
    )
    query.append(orm.Computer, with_node='calc', tag='computer', project=['uuid'], filters=filters_computer)
    query.append(orm.User, with_node='calc', filters={'email': user.email})

    if query.count() == 0:
        return None

    path_mapping = {}

    for remote_data, computer_uuid in query.iterall():
        path_mapping.setdefault(computer_uuid, []).append(remote_data)

    return path_mapping



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Utility functions to operate on datetime objects."""

from datetime import datetime, timedelta, timezone, tzinfo
from typing import Optional

utc = timezone.utc  # Simply forward this attribute from the :mod:`datetime.timezone` built in library


def now() -> datetime:
    """Return the datetime object of the current time.

    :return: datetime object represeting current time
    """
    return datetime.now().astimezone()


def make_aware(value: datetime, tz: Optional[tzinfo] = None) -> datetime:
    """Make the given datetime object timezone aware.

    :param value: The datetime object to make aware.
    :param tz: The timezone to set. If not defined the system local timezone is assumed for the target timezone.
    :return: A timezone aware datetime object.
    """
    return value.astimezone(tz)


def localtime(value: datetime) -> datetime:
    """Make a :class:`datetime.datetime` object timezone aware with the local timezone.

    :param value: The datetime object to make aware.
    :return: A timezone aware datetime object with the timezone set to that of the operating system.
    """
    return make_aware(value)


def timezone_from_name(name: str) -> tzinfo:
    """Return a :class:`datetime.tzinfo` instance corresponding to the given timezone name.

    :param name: The timezone name. Should correspond to a known name in the Olsen database.
        https://en.wikipedia.org/wiki/Tz_database
    :returns: The corresponding timezone object.
    :raises ValueError: if the timezone name is unknown.
    """
    import pytz

    try:
        return pytz.timezone(name)
    except pytz.exceptions.UnknownTimeZoneError as exception:
        raise ValueError(f'unknown timezone: {name}') from exception


def delta(from_time: datetime, to_time: Optional[datetime] = None) -> timedelta:
    """Return the datetime object representing the different between two datetime objects.

    :param from_time: The starting datetime object.
    :param to_time: The end datetime object. If not specified :func:`aiida.common.timezone.now` is used.
    :return: The delta datetime object.
    """
    return make_aware(to_time or now()) - make_aware(from_time)



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Data plugin representing an executable code stored in AiiDA's storage.

This plugin should be used for executables that are not already installed on the target computer, but instead are
available on the machine where AiiDA is running. The plugin assumes that the code is self-contained by a single
directory containing all the necessary files, including a main executable. When constructing a ``PortableCode``, passing
the absolute filepath as ``filepath_files`` will make sure that all the files contained within are uploaded to AiiDA's
storage. The ``filepath_executable`` should indicate the filename of the executable within that directory. Each time a
:class:`aiida.engine.CalcJob` is run using a ``PortableCode``, the uploaded files will be automatically copied to the
working directory on the selected computer and the executable will be run there.
"""

from __future__ import annotations

import pathlib
import typing as t

from pydantic import field_validator

from aiida.common import exceptions
from aiida.common.folders import Folder
from aiida.common.lang import type_check
from aiida.common.pydantic import MetadataField
from aiida.orm import Computer

from .abstract import AbstractCode
from .legacy import Code

__all__ = ('PortableCode',)


class PortableCode(Code):
    """Data plugin representing an executable code stored in AiiDA's storage."""

    _EMIT_CODE_DEPRECATION_WARNING: bool = False
    _KEY_ATTRIBUTE_FILEPATH_EXECUTABLE: str = 'filepath_executable'

    class Model(AbstractCode.Model):
        """Model describing required information to create an instance."""

        filepath_files: t.Union[str, pathlib.Path] = MetadataField(
            ...,
            title='Code directory',
            description='Filepath to directory containing code files.',
            short_name='-F',
            priority=2,
        )
        filepath_executable: str = MetadataField(
            ...,
            title='Filepath executable',
            description='Relative filepath of executable with directory of code files.',
            short_name='-X',
            priority=1,
        )

        @field_validator('filepath_files')
        @classmethod
        def validate_filepath_files(cls, value: str) -> pathlib.Path:
            """Validate that ``filepath_files`` is an existing directory."""
            filepath = pathlib.Path(value)
            if not filepath.exists():
                raise ValueError(f'The filepath `{value}` does not exist.')
            if not filepath.is_dir():
                raise ValueError(f'The filepath `{value}` is not a directory.')
            return filepath

    def __init__(self, filepath_executable: str, filepath_files: pathlib.Path, **kwargs):
        """Construct a new instance.

        .. note:: If the files necessary for this code are not all located in a single directory or the directory
            contains files that should not be uploaded, and so the ``filepath_files`` cannot be used. One can use the
            methods of the :class:`aiida.orm.nodes.repository.NodeRepository` class. This can be accessed through the
            ``base.repository`` attribute of the instance after it has been constructed. For example::

                code = PortableCode(filepath_executable='some_name.exe')
                code.put_object_from_file()
                code.put_object_from_filelike()
                code.put_object_from_tree()

        :param filepath_executable: The relative filepath of the executable within the directory of uploaded files.
        :param filepath_files: The filepath to the directory containing all the files of the code.
        """
        super().__init__(**kwargs)
        type_check(filepath_files, pathlib.Path)
        self.filepath_executable = filepath_executable  # type: ignore[assignment]
        self.base.repository.put_object_from_tree(str(filepath_files))

    def _validate(self):
        """Validate the instance by checking that an executable is defined and it is part of the repository files.

        :raises :class:`aiida.common.exceptions.ValidationError`: If the state of the node is invalid.
        """
        super(Code, self)._validate()  # Change to ``super()._validate()`` once deprecated ``Code`` class is removed.

        try:
            filepath_executable = self.filepath_executable
        except TypeError as exception:
            raise exceptions.ValidationError('The `filepath_executable` is not set.') from exception

        objects = self.base.repository.list_object_names()

        if str(filepath_executable) not in objects:
            raise exceptions.ValidationError(
                f'The executable `{filepath_executable}` is not one of the uploaded files: {objects}'
            )

    def can_run_on_computer(self, computer: Computer) -> bool:
        """Return whether the code can run on a given computer.

        A ``PortableCode`` should be able to be run on any computer in principle.

        :param computer: The computer.
        :return: ``True`` if the provided computer is the same as the one configured for this code.
        """
        return True

    def get_executable(self) -> pathlib.PurePosixPath:
        """Return the executable that the submission script should execute to run the code.

        :return: The executable to be called in the submission script.
        """
        return self.filepath_executable

    def validate_working_directory(self, folder: Folder):
        """Validate content of the working directory created by the :class:`~aiida.engine.CalcJob` plugin.

        This method will be called by :meth:`~aiida.engine.processes.calcjobs.calcjob.CalcJob.presubmit` when a new
        calculation job is launched, passing the :class:`~aiida.common.folders.Folder` that was used by the plugin used
        for the calculation to create the input files for the working directory. This method can be overridden by
        implementations of the ``AbstractCode`` class that need to validate the contents of that folder.

        :param folder: A sandbox folder that the ``CalcJob`` plugin wrote input files to that will be copied to the
            working directory for the corresponding calculation job instance.
        :raises PluginInternalError: The ``CalcJob`` plugin created a file that has the same relative filepath as the
            executable for this portable code.
        """
        if str(self.filepath_executable) in folder.get_content_list():
            raise exceptions.PluginInternalError(
                f'The plugin created a file {self.filepath_executable} that is also the executable name!'
            )

    @property
    def full_label(self) -> str:
        """Return the full label of this code.

        The full label can be just the label itself but it can be something else. However, it at the very least has to
        include the label of the code.

        :return: The full label of the code.
        """
        return self.label

    @property
    def filepath_executable(self) -> pathlib.PurePosixPath:
        """Return the relative filepath of the executable that this code represents.

        :return: The relative filepath of the executable.
        """
        return pathlib.PurePosixPath(self.base.attributes.get(self._KEY_ATTRIBUTE_FILEPATH_EXECUTABLE))

    @filepath_executable.setter
    def filepath_executable(self, value: str) -> None:
        """Set the relative filepath of the executable that this code represents.

        :param value: The relative filepath of the executable within the directory of uploaded files.
        """
        type_check(value, str)

        if pathlib.PurePosixPath(value).is_absolute():
            raise ValueError('The `filepath_executable` should not be absolute.')

        self.base.attributes.set(self._KEY_ATTRIBUTE_FILEPATH_EXECUTABLE, value)



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Client to interact with the daemon."""

from __future__ import annotations

import contextlib
import enum
import os
import pathlib
import shutil
import socket
import subprocess
import sys
import tempfile
import time
import typing as t
from typing import TYPE_CHECKING

import psutil

from aiida.common.exceptions import AiidaException, ConfigurationError
from aiida.common.lang import type_check
from aiida.common.log import AIIDA_LOGGER
from aiida.manage.configuration import get_config, get_config_option
from aiida.manage.configuration.profile import Profile
from aiida.manage.manager import get_manager

if TYPE_CHECKING:
    from circus.client import CircusClient

LOGGER = AIIDA_LOGGER.getChild('engine.daemon.client')

VERDI_BIN = shutil.which('verdi')
# Recent versions of virtualenv create the environment variable VIRTUAL_ENV
VIRTUALENV = os.environ.get('VIRTUAL_ENV', None)

__all__ = ('DaemonClient', 'get_daemon_client')


class ControllerProtocol(enum.Enum):
    """The protocol to use for the controller of the Circus daemon."""

    IPC = 0
    TCP = 1


class DaemonException(AiidaException):
    """Base class for exceptions related to the daemon."""


class DaemonNotRunningException(DaemonException):
    """Raised when a connection to the daemon is attempted but it is not running."""


class DaemonTimeoutException(DaemonException):
    """Raised when a connection to the daemon is attempted but it times out."""


class DaemonStalePidException(DaemonException):
    """Raised when a connection to the daemon is attempted but it fails and the PID file appears to be stale."""


def get_daemon_client(profile_name: str | None = None) -> 'DaemonClient':
    """Return the daemon client for the given profile or the current profile if not specified.

    :param profile_name: Optional profile name to load.
    :return: The daemon client.

    :raises aiida.common.MissingConfigurationError: if the configuration file cannot be found.
    :raises aiida.common.ProfileConfigurationError: if the given profile does not exist.
    """
    profile = get_manager().load_profile(profile_name)
    return DaemonClient(profile)


class DaemonClient:
    """Client to interact with the daemon."""

    _DAEMON_NAME = 'aiida-{name}'
    _ENDPOINT_PROTOCOL = ControllerProtocol.IPC

    def __init__(self, profile: Profile):
        """Construct an instance for a given profile.

        :param profile: The profile instance.
        """
        type_check(profile, Profile)
        config = get_config()
        self._profile = profile
        self._socket_directory: str | None = None
        self._daemon_timeout: int = config.get_option('daemon.timeout', scope=profile.name)

        if self._profile.process_control_backend is None:
            raise ConfigurationError(
                f'profile `{self._profile.name}` does not define a broker so the daemon cannot be used.'
            )

    @property
    def profile(self) -> Profile:
        return self._profile

    @property
    def daemon_name(self) -> str:
        """Get the daemon name which is tied to the profile name."""
        return self._DAEMON_NAME.format(name=self.profile.name)

    @property
    def _verdi_bin(self) -> str:
        """Return the absolute path to the ``verdi`` binary.

        :raises ConfigurationError: If the path to ``verdi`` could not be found
        """
        if VERDI_BIN is None:
            raise ConfigurationError(
                "Unable to find 'verdi' in the path. Make sure that you are working "
                "in a virtual environment, or that at least the 'verdi' executable is on the PATH"
            )

        return VERDI_BIN

    def cmd_start_daemon(self, number_workers: int = 1, foreground: bool = False) -> list[str]:
        """Return the command to start the daemon.

        :param number_workers: Number of daemon workers to start.
        :param foreground: Whether to launch the subprocess in the background or not.
        """
        command = [self._verdi_bin, '-p', self.profile.name, 'daemon', 'start-circus', str(number_workers)]

        if foreground:
            command.append('--foreground')

        return command

    @property
    def cmd_start_daemon_worker(self) -> list[str]:
        """Return the command to start a daemon worker process."""
        return [self._verdi_bin, '-p', self.profile.name, 'daemon', 'worker']

    @property
    def loglevel(self) -> str:
        return get_config_option('logging.circus_loglevel')

    @property
    def virtualenv(self) -> str | None:
        return VIRTUALENV

    @property
    def circus_log_file(self) -> str:
        return self.profile.filepaths['circus']['log']

    @property
    def circus_pid_file(self) -> str:
        return self.profile.filepaths['circus']['pid']

    @property
    def circus_port_file(self) -> str:
        return self.profile.filepaths['circus']['port']

    @property
    def circus_socket_file(self) -> str:
        return self.profile.filepaths['circus']['socket']['file']

    @property
    def circus_socket_endpoints(self) -> dict[str, str]:
        return self.profile.filepaths['circus']['socket']

    @property
    def daemon_log_file(self) -> str:
        return self.profile.filepaths['daemon']['log']

    @property
    def daemon_pid_file(self) -> str:
        return self.profile.filepaths['daemon']['pid']

    def get_circus_port(self) -> int:
        """Retrieve the port for the circus controller, which should be written to the circus port file.

        If the daemon is running, the port file should exist and contain the port to which the controller is connected.
        If it cannot be read, a RuntimeError will be thrown. If the daemon is not running, an available port will be
        requested from the operating system, written to the port file and returned.

        :return: The port for the circus controller.
        """
        if self.is_daemon_running:
            try:
                with open(self.circus_port_file, 'r', encoding='utf8') as fhandle:
                    return int(fhandle.read().strip())
            except (ValueError, OSError):
                raise RuntimeError('daemon is running so port file should have been there but could not read it')
        else:
            port = self.get_available_port()
            with open(self.circus_port_file, 'w', encoding='utf8') as fhandle:
                fhandle.write(str(port))

            return port

    @staticmethod
    def get_env() -> dict[str, str]:
        """Return the environment for this current process.

        This method is used to pass variables from the environment of the current process to a subprocess that is
        spawned when the daemon or a daemon worker is started.

        It replicates the ``PATH``, ``PYTHONPATH` and the ``AIIDA_PATH`` environment variables. The ``PYTHONPATH``
        variable ensures that all Python modules that can be imported by the parent process, are also importable by
        the subprocess. The ``AIIDA_PATH`` variable ensures that the subprocess will use the same AiiDA configuration
        directory as used by the current process.
        """
        env = os.environ.copy()
        env['PATH'] = ':'.join([os.path.dirname(sys.executable), env['PATH']])
        env['PYTHONPATH'] = ':'.join(sys.path)
        env['AIIDA_PATH'] = get_config().dirpath
        env['PYTHONUNBUFFERED'] = 'True'
        return env

    def get_circus_socket_directory(self) -> str:
        """Retrieve the absolute path of the directory where the circus sockets are stored.

        If the daemon is running, the sockets file should exist and contain the absolute path of the directory that
        contains the sockets of the circus endpoints. If it cannot be read, a ``RuntimeError`` will be thrown. If the
        daemon is not running, a temporary directory will be created and its path will be written to the sockets file
        and returned.

        .. note:: A temporary folder needs to be used for the sockets because UNIX limits the filepath length to
            107 bytes. Placing the socket files in the AiiDA config folder might seem like the more logical choice
            but that folder can be placed in an arbitrarily nested directory, the socket filename will exceed the
            limit. The solution is therefore to always store them in the temporary directory of the operation system
            whose base path is typically short enough as to not exceed the limit

        :return: The absolute path of directory to write the sockets to.
        """
        if self.is_daemon_running:
            try:
                with open(self.circus_socket_file, 'r', encoding='utf8') as fhandle:
                    content = fhandle.read().strip()
                return content
            except (ValueError, OSError):
                raise RuntimeError('daemon is running so sockets file should have been there but could not read it')
        else:
            # The SOCKET_DIRECTORY is already set, a temporary directory was already created and the same should be used
            if self._socket_directory is not None:
                return self._socket_directory

            socket_dir_path = tempfile.mkdtemp()
            with open(self.circus_socket_file, 'w', encoding='utf8') as fhandle:
                fhandle.write(str(socket_dir_path))

            self._socket_directory = socket_dir_path
            return socket_dir_path

    def get_daemon_pid(self) -> int | None:
        """Get the daemon pid which should be written in the daemon pid file specific to the profile.

        :return: The pid of the circus daemon process or None if not found.
        """
        if os.path.isfile(self.circus_pid_file):
            try:
                with open(self.circus_pid_file, 'r', encoding='utf8') as fhandle:
                    content = fhandle.read().strip()
                return int(content)
            except (ValueError, OSError):
                return None
        else:
            return None

    @property
    def is_daemon_running(self) -> bool:
        """Return whether the daemon is running, which is determined by seeing if the daemon pid file is present.

        :return: True if daemon is running, False otherwise.
        """
        return self.get_daemon_pid() is not None

    def delete_circus_socket_directory(self) -> None:
        """Attempt to delete the directory used to store the circus endpoint sockets.

        Will not raise if the directory does not exist.
        """
        directory = self.get_circus_socket_directory()

        try:
            shutil.rmtree(directory)
        except OSError as exception:
            if exception.errno == 2:
                pass
            else:
                raise

    @classmethod
    def get_available_port(cls):
        """Get an available port from the operating system.

        :return: A currently available port.
        """
        open_socket = socket.socket()
        open_socket.bind(('', 0))
        return open_socket.getsockname()[1]

    def get_controller_endpoint(self):
        """Get the endpoint string for the circus controller.

        For the IPC protocol a profile specific socket will be used, whereas for the TCP protocol an available port will
        be found and saved in the profile specific port file.

        :return: The endpoint string.
        """
        if self._ENDPOINT_PROTOCOL == ControllerProtocol.IPC:
            endpoint = self.get_ipc_endpoint('controller')
        elif self._ENDPOINT_PROTOCOL == ControllerProtocol.TCP:
            endpoint = self.get_tcp_endpoint(self.get_circus_port())
        else:
            raise ValueError(f'invalid controller protocol {self._ENDPOINT_PROTOCOL}')

        return endpoint

    def get_pubsub_endpoint(self):
        """Get the endpoint string for the circus pubsub endpoint.

        For the IPC protocol a profile specific socket will be used, whereas for the TCP protocol any available port
        will be used.

        :return: The endpoint string.
        """
        if self._ENDPOINT_PROTOCOL == ControllerProtocol.IPC:
            endpoint = self.get_ipc_endpoint('pubsub')
        elif self._ENDPOINT_PROTOCOL == ControllerProtocol.TCP:
            endpoint = self.get_tcp_endpoint()
        else:
            raise ValueError(f'invalid controller protocol {self._ENDPOINT_PROTOCOL}')

        return endpoint

    def get_stats_endpoint(self):
        """Get the endpoint string for the circus stats endpoint.

        For the IPC protocol a profile specific socket will be used, whereas for the TCP protocol any available port
        will be used.

        :return: The endpoint string.
        """
        if self._ENDPOINT_PROTOCOL == ControllerProtocol.IPC:
            endpoint = self.get_ipc_endpoint('stats')
        elif self._ENDPOINT_PROTOCOL == ControllerProtocol.TCP:
            endpoint = self.get_tcp_endpoint()
        else:
            raise ValueError(f'invalid controller protocol {self._ENDPOINT_PROTOCOL}')

        return endpoint

    def get_ipc_endpoint(self, endpoint):
        """Get the ipc endpoint string for a circus daemon endpoint for a given socket.

        :param endpoint: The circus endpoint for which to return a socket.
        :return: The ipc endpoint string.
        """
        filepath = self.get_circus_socket_directory()
        filename = self.circus_socket_endpoints[endpoint]
        template = 'ipc://{filepath}/{filename}'
        endpoint = template.format(filepath=filepath, filename=filename)

        return endpoint

    def get_tcp_endpoint(self, port=None):
        """Get the tcp endpoint string for a circus daemon endpoint.

        If the port is unspecified, the operating system will be asked for a currently available port.

        :param port: A port to use for the endpoint.
        :return: The tcp endpoint string.
        """
        if port is None:
            port = self.get_available_port()

        template = 'tcp://127.0.0.1:{port}'
        endpoint = template.format(port=port)

        return endpoint

    @contextlib.contextmanager
    def get_client(self, timeout: int | None = None) -> 'CircusClient':
        """Return an instance of the CircusClient.

        The endpoint is defined by the controller endpoint, which used the port that was written to the port file upon
        starting of the daemon.

        :param timeout: Optional timeout to set for trying to reach the circus daemon. Default is set on the client upon
            instantiation taken from the ``daemon.timeout`` config option.
        :return: CircusClient instance
        """
        from circus.client import CircusClient

        try:
            client = CircusClient(endpoint=self.get_controller_endpoint(), timeout=timeout or self._daemon_timeout)
            yield client
        finally:
            client.stop()

    def call_client(self, command: dict[str, t.Any], timeout: int | None = None) -> dict[str, t.Any]:
        """Call the client with a specific command.

        Will check whether the daemon is running first by checking for the pid file. When the pid is found yet the call
        still fails with a timeout, this means the daemon was actually not running and it was terminated unexpectedly
        causing the pid file to not be cleaned up properly.

        :param command: Command to call the circus client with.
        :param timeout: Optional timeout to set for trying to reach the circus daemon. Default is set on the client upon
            instantiation taken from the ``daemon.timeout`` config option.
        :return: The result of the circus client call.
        :raises DaemonException: If the daemon is not running or cannot be reached.
        :raises DaemonTimeoutException: If the connection to the daemon timed out.
        :raises DaemonException: If the connection to the daemon failed for any other reason.
        """
        from circus.exc import CallError

        try:
            with self.get_client(timeout=timeout) as client:
                result = client.call(command)
        except CallError as exception:
            if self.get_daemon_pid() is None:
                raise DaemonNotRunningException('The daemon is not running.') from exception

            if self._is_pid_file_stale:
                raise DaemonStalePidException(
                    'The daemon could not be reached, seemingly because of a stale PID file. Either stop or start the '
                    'daemon to remove it and restore the daemon to a functional state.'
                ) from exception

            if str(exception) == 'Timed out.':
                raise DaemonTimeoutException('Connection to the daemon timed out.') from exception

            raise DaemonException('Connection to the daemon failed.') from exception

        return result

    def get_status(self, timeout: int | None = None) -> dict[str, t.Any]:
        """Return the status of the daemon.

        :param timeout: Optional timeout to set for trying to reach the circus daemon. Default is set on the client upon
            instantiation taken from the ``daemon.timeout`` config option.
        :returns: The client call response. If successful, will contain 'pid' key.
        """
        command = {'command': 'status', 'properties': {'name': self.daemon_name}}
        response = self.call_client(command, timeout=timeout)
        response['pid'] = self.get_daemon_pid()
        return response

    def get_numprocesses(self, timeout: int | None = None) -> dict[str, t.Any]:
        """Get the number of running daemon processes.

        :param timeout: Optional timeout to set for trying to reach the circus daemon. Default is set on the client upon
            instantiation taken from the ``daemon.timeout`` config option.
        :return: The client call response. If successful, will contain 'numprocesses' key.
        """
        command = {'command': 'numprocesses', 'properties': {'name': self.daemon_name}}
        return self.call_client(command, timeout=timeout)

    def get_worker_info(self, timeout: int | None = None) -> dict[str, t.Any]:
        """Get workers statistics for this daemon.

        :param timeout: Optional timeout to set for trying to reach the circus daemon. Default is set on the client upon
            instantiation taken from the ``daemon.timeout`` config option.
        :return: The client call response. If successful, will contain 'info' key.
        """
        command = {'command': 'stats', 'properties': {'name': self.daemon_name}}
        return self.call_client(command, timeout=timeout)

    def get_daemon_info(self, timeout: int | None = None) -> dict[str, t.Any]:
        """Get statistics about this daemon itself.

        :param timeout: Optional timeout to set for trying to reach the circus daemon. Default is set on the client upon
            instantiation taken from the ``daemon.timeout`` config option.
        :return: The client call response. If successful, will contain 'info' key.
        """
        command = {'command': 'dstats', 'properties': {}}
        return self.call_client(command, timeout=timeout)

    def increase_workers(self, number: int, timeout: int | None = None) -> dict[str, t.Any]:
        """Increase the number of workers.

        :param number: The number of workers to add.
        :param timeout: Optional timeout to set for trying to reach the circus daemon. Default is set on the client upon
            instantiation taken from the ``daemon.timeout`` config option.
        :return: The client call response.
        """
        command = {'command': 'incr', 'properties': {'name': self.daemon_name, 'nb': number}}
        return self.call_client(command, timeout=timeout)

    def decrease_workers(self, number: int, timeout: int | None = None) -> dict[str, t.Any]:
        """Decrease the number of workers.

        :param number: The number of workers to remove.
        :param timeout: Optional timeout to set for trying to reach the circus daemon. Default is set on the client upon
            instantiation taken from the ``daemon.timeout`` config option.
        :return: The client call response.
        """
        command = {'command': 'decr', 'properties': {'name': self.daemon_name, 'nb': number}}
        return self.call_client(command, timeout=timeout)

    def start_daemon(
        self, number_workers: int = 1, foreground: bool = False, wait: bool = True, timeout: int | None = None
    ) -> None:
        """Start the daemon in a sub process running in the background.

        :param number_workers: Number of daemon workers to start.
        :param foreground: Whether to launch the subprocess in the background or not.
        :param wait: Boolean to indicate whether to wait for the result of the command.
        :param timeout: Optional timeout to set for trying to reach the circus daemon after the subprocess has started.
            Default is set on the client upon instantiation taken from the ``daemon.timeout`` config option.
        :raises DaemonException: If the command to start the daemon subprocess excepts.
        :raises DaemonTimeoutException: If the daemon starts but then is unresponsive or in an unexpected state.
        """
        self._clean_potentially_stale_pid_file()

        env = self.get_env()
        command = self.cmd_start_daemon(number_workers, foreground)
        timeout = timeout or self._daemon_timeout

        try:
            subprocess.check_output(command, env=env, stderr=subprocess.STDOUT)
        except subprocess.CalledProcessError as exception:
            raise DaemonException('The daemon failed to start.') from exception

        if not wait:
            return

        self._await_condition(
            lambda: self.is_daemon_running,
            DaemonTimeoutException(f'The daemon failed to start or is unresponsive after {timeout} seconds.'),
            timeout=timeout,
        )

    def restart_daemon(self, wait: bool = True, timeout: int | None = None) -> dict[str, t.Any]:
        """Restart the daemon.

        :param wait: Boolean to indicate whether to wait for the result of the command.
        :param timeout: Optional timeout to set for trying to reach the circus daemon. Default is set on the client upon
            instantiation taken from the ``daemon.timeout`` config option.
        :returns: The client call response.
        :raises DaemonException: If the daemon is not running or cannot be reached.
        :raises DaemonTimeoutException: If the connection to the daemon timed out.
        :raises DaemonException: If the connection to the daemon failed for any other reason.
        """
        command = {'command': 'restart', 'properties': {'name': self.daemon_name, 'waiting': wait}}
        return self.call_client(command, timeout=timeout)

    def stop_daemon(self, wait: bool = True, timeout: int | None = None) -> dict[str, t.Any]:
        """Stop the daemon.

        :param wait: Boolean to indicate whether to wait for the result of the command.
        :param timeout: Optional timeout to set for trying to reach the circus daemon. Default is set on the client upon
            instantiation taken from the ``daemon.timeout`` config option.
        :returns: The client call response.
        :raises DaemonException: If the daemon is not running or cannot be reached.
        :raises DaemonTimeoutException: If the connection to the daemon timed out.
        :raises DaemonException: If the connection to the daemon failed for any other reason.
        """
        self._clean_potentially_stale_pid_file()

        command = {'command': 'quit', 'properties': {'waiting': wait}}
        response = self.call_client(command, timeout=timeout)

        if self._ENDPOINT_PROTOCOL == ControllerProtocol.IPC:
            self.delete_circus_socket_directory()

        return response

    def _clean_potentially_stale_pid_file(self) -> None:
        """Check the daemon PID file and delete it if it is likely to be stale."""
        try:
            self._check_pid_file()
        except DaemonException as exception:
            pathlib.Path(self.circus_pid_file).unlink(missing_ok=True)
            LOGGER.warning(f'Deleted apparently stale daemon PID file: {exception}')

    @property
    def _is_pid_file_stale(self) -> bool:
        """Return whether the daemon PID file is likely to be stale.

        :returns: ``True`` if the PID file is likely to be stale, ``False`` otherwise.
        """
        try:
            self._check_pid_file()
        except DaemonException:
            return True

        return False

    def _check_pid_file(self) -> None:
        """Check that the daemon's PID file is not stale.

        Checks if the PID contained in the circus PID file matches a valid running ``verdi`` process. The PID file is
        considered stale if any of the following conditions are true:

        * The process with the given PID no longer exists
        * The process name does not match the command of the circus daemon
        * The process username does not match the username of this Python interpreter

        In the latter two cases, the process with the PID of the PID file exists, but it is very likely that it is not
        the original process that created the PID file, since the command or user is different, indicating the original
        process died and the PID was recycled for a new process.

        The PID file can got stale if a system is shut down suddenly and so the process is killed but the PID file is
        not deleted in time. When the `get_daemon_pid()` method is called, an incorrect PID is returned. Alternatively,
        another process or the user may have meddled with the PID file in some way, corrupting it.

        :raises DaemonException: If the PID file is likely to be stale.
        """
        pid = self.get_daemon_pid()

        if pid is None:
            return

        try:
            process = psutil.Process(pid)

            # The circus daemon process can appear as ``start-circus`` or ``circusd``. See this issue comment for
            # details: https://github.com/aiidateam/aiida-core/issues/5336#issuecomment-1376093322
            if not any(cmd in process.cmdline() for cmd in ['start-circus', 'circusd']):
                raise DaemonException(
                    f'process command `{process.cmdline()}` of PID `{pid}` does not match expected AiiDA daemon command'
                )

            process_user = process.username()
            current_user = psutil.Process().username()

            if process_user != current_user:
                raise DaemonException(
                    f'process user `{process_user}` of PID `{pid}` does not match current user `{current_user}`'
                )

        except (psutil.AccessDenied, psutil.NoSuchProcess, DaemonException) as exception:
            raise DaemonException(exception) from exception

    @staticmethod
    def _await_condition(condition: t.Callable, exception: Exception, timeout: int = 5, interval: float = 0.1):
        """Await a condition to evaluate to ``True`` or raise the exception if the timeout is reached.

        :param condition: A callable that is waited for to return ``True``.
        :param exception: Raise this exception if ``condition`` does not return ``True`` after ``timeout`` seconds.
        :param timeout: Wait this number of seconds for ``condition`` to return ``True`` before raising.
        :param interval: The time in seconds to wait between invocations of ``condition``.
        :raises: The exception provided by ``exception`` if timeout is reached.
        """
        start_time = time.time()

        while not condition():
            time.sleep(interval)

            if time.time() - start_time > timeout:
                raise exception

    def _start_daemon(self, number_workers: int = 1, foreground: bool = False) -> None:
        """Start the daemon.

        .. warning:: This will daemonize the current process and put it in the background. It is most likely not what
            you want to call if you want to start the daemon from the Python API. Instead you probably will want to use
            the :meth:`aiida.engine.daemon.client.DaemonClient.start_daemon` function instead.

        :param number_workers: Number of daemon workers to start.
        :param foreground: Whether to launch the subprocess in the background or not.
        """
        from circus import get_arbiter
        from circus import logger as circus_logger
        from circus.circusd import daemonize
        from circus.pidfile import Pidfile
        from circus.util import check_future_exception_and_log, configure_logger

        if foreground and number_workers > 1:
            raise ValueError('can only run a single worker when running in the foreground')

        loglevel = self.loglevel
        logoutput = '-'

        if not foreground:
            logoutput = self.circus_log_file

        arbiter_config = {
            'controller': self.get_controller_endpoint(),
            'pubsub_endpoint': self.get_pubsub_endpoint(),
            'stats_endpoint': self.get_stats_endpoint(),
            'logoutput': logoutput,
            'loglevel': loglevel,
            'debug': False,
            'statsd': True,
            'pidfile': self.circus_pid_file,
            'watchers': [
                {
                    'cmd': ' '.join(self.cmd_start_daemon_worker),
                    'name': self.daemon_name,
                    'numprocesses': number_workers,
                    'virtualenv': self.virtualenv,
                    'copy_env': True,
                    'stdout_stream': {
                        'class': 'FileStream',
                        'filename': self.daemon_log_file,
                    },
                    'stderr_stream': {
                        'class': 'FileStream',
                        'filename': self.daemon_log_file,
                    },
                    'env': self.get_env(),
                }
            ],
        }

        if not foreground:
            daemonize()

        arbiter = get_arbiter(**arbiter_config)
        pidfile = Pidfile(arbiter.pidfile)
        pidfile.create(os.getpid())

        # Configure the logger
        loggerconfig = arbiter.loggerconfig or None
        configure_logger(circus_logger, loglevel, logoutput, loggerconfig)

        # Main loop
        should_restart = True

        while should_restart:
            try:
                future = arbiter.start()
                should_restart = False
                if check_future_exception_and_log(future) is None:
                    should_restart = arbiter._restarting
            except Exception as exception:
                # Emergency stop
                arbiter.loop.run_sync(arbiter._emergency_stop)
                raise exception
            except KeyboardInterrupt:
                pass
            finally:
                arbiter = None
                if pidfile is not None:
                    pidfile.unlink()



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Provides tools for visualization of the provenance graph."""

# AUTO-GENERATED

# fmt: off

from .graph import *

__all__ = (
    'Graph',
    'default_link_styles',
    'default_node_styles',
    'default_node_sublabels',
    'pstate_node_styles',
)

# fmt: on



"""Interface for a message broker that facilitates communication with and between process runners."""

import abc
import typing as t

if t.TYPE_CHECKING:
    from aiida.manage.configuration.profile import Profile

__all__ = ('Broker',)


class Broker:
    """Interface for a message broker that facilitates communication with and between process runners."""

    def __init__(self, profile: 'Profile') -> None:
        """Construct a new instance.

        :param profile: The profile.
        """
        self._profile = profile

    @abc.abstractmethod
    def get_communicator(self):
        """Return an instance of :class:`kiwipy.Communicator`."""

    @abc.abstractmethod
    def iterate_tasks(self):
        """Return an iterator over the tasks in the launch queue."""

    @abc.abstractmethod
    def close(self):
        """Close the broker."""



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Implementation of ORM backend entities."""

import warnings

from sqlalchemy.exc import SAWarning

# The :meth:`sqlalchemy.orm.unitofwork.UOWTransaction.register_object` method emits a warning whenever an object is
# registered that is not part of the session. This can happen when the session is committed or flushed and an object
# inside the session contains a reference to another object, for example through a relationship, is not explicitly part
# of the session. If that referenced object is not already stored and persisted, it might get lost. On the other hand,
# if the object was already persisted before, there is no risk.
#
# This situation occurs a lot in AiiDA's code base. Prime example is when a new process is created. Typically the input
# nodes are either already stored, or stored first. As soon as they get stored, the session is committed and the session
# is reset by expiring all objects. Now, the input links are created from the input nodes to the process node, and at
# the end the process node is stored to commit and persist it with the links. It is at this point that Sqlalchemy
# realises that the input nodes are not explicitly part of the session.
#
# One direct solution would be to add the input nodes again to the session before committing the process node and the
# links. However, this code is part of the backend independent :mod:`aiida.orm` module and this is a backend-specific
# problem. This is also just one example and there are most likely other places in the code where the problem arises.
# Therefore, as a work around, the specific warning is ignored through a warrning filter.
warnings.filterwarnings(
    'ignore', category=SAWarning, message='Object of type .* not in session, .* operation along .* will not proceed'
)



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Abstract BackendNode and BackendNodeCollection implementation."""

import abc
from datetime import datetime
from typing import TYPE_CHECKING, Any, Dict, Iterable, List, Optional, Sequence, Tuple, TypeVar

from .entities import BackendCollection, BackendEntity, BackendEntityExtrasMixin

if TYPE_CHECKING:
    from ..utils import LinkTriple
    from .computers import BackendComputer
    from .users import BackendUser

__all__ = ('BackendNode', 'BackendNodeCollection')

BackendNodeType = TypeVar('BackendNodeType', bound='BackendNode')


class BackendNode(BackendEntity, BackendEntityExtrasMixin, metaclass=abc.ABCMeta):
    """Backend implementation for the `Node` ORM class.

    A node stores data input or output from a computation.
    """

    @abc.abstractmethod
    def clone(self: BackendNodeType) -> BackendNodeType:
        """Return an unstored clone of ourselves.

        :return: an unstored `BackendNode` with the exact same attributes and extras as self
        """

    @property
    @abc.abstractmethod
    def uuid(self) -> str:
        """Return the node UUID.

        :return: the string representation of the UUID
        """

    @property
    @abc.abstractmethod
    def node_type(self) -> str:
        """Return the node type.

        :return: the node type
        """

    @property
    @abc.abstractmethod
    def process_type(self) -> Optional[str]:
        """Return the node process type.

        :return: the process type
        """

    @process_type.setter
    @abc.abstractmethod
    def process_type(self, value: Optional[str]) -> None:
        """Set the process type.

        :param value: the new value to set
        """

    @property
    @abc.abstractmethod
    def label(self) -> str:
        """Return the node label.

        :return: the label
        """

    @label.setter
    @abc.abstractmethod
    def label(self, value: str) -> None:
        """Set the label.

        :param value: the new value to set
        """

    @property
    @abc.abstractmethod
    def description(self) -> str:
        """Return the node description.

        :return: the description
        """

    @description.setter
    @abc.abstractmethod
    def description(self, value: str) -> None:
        """Set the description.

        :param value: the new value to set
        """

    @property
    @abc.abstractmethod
    def repository_metadata(self) -> Dict[str, Any]:
        """Return the node repository metadata.

        :return: the repository metadata
        """

    @repository_metadata.setter
    @abc.abstractmethod
    def repository_metadata(self, value: Dict[str, Any]) -> None:
        """Set the repository metadata.

        :param value: the new value to set
        """

    @property
    @abc.abstractmethod
    def computer(self) -> Optional['BackendComputer']:
        """Return the computer of this node.

        :return: the computer or None
        """

    @computer.setter
    @abc.abstractmethod
    def computer(self, computer: Optional['BackendComputer']) -> None:
        """Set the computer of this node.

        :param computer: a `BackendComputer`
        """

    @property
    @abc.abstractmethod
    def user(self) -> 'BackendUser':
        """Return the user of this node.

        :return: the user
        """

    @user.setter
    @abc.abstractmethod
    def user(self, user: 'BackendUser') -> None:
        """Set the user of this node.

        :param user: a `BackendUser`
        """

    @property
    @abc.abstractmethod
    def ctime(self) -> datetime:
        """Return the node ctime.

        :return: the ctime
        """

    @property
    @abc.abstractmethod
    def mtime(self) -> datetime:
        """Return the node mtime.

        :return: the mtime
        """

    @abc.abstractmethod
    def add_incoming(self, source: 'BackendNode', link_type, link_label):
        """Add a link of the given type from a given node to ourself.

        :param source: the node from which the link is coming
        :param link_type: the link type
        :param link_label: the link label
        :return: True if the proposed link is allowed, False otherwise
        :raise TypeError: if `source` is not a Node instance or `link_type` is not a `LinkType` enum
        :raise ValueError: if the proposed link is invalid
        :raise aiida.common.ModificationNotAllowed: if either source or target node is not stored
        """

    @abc.abstractmethod
    def store(
        self: BackendNodeType, links: Optional[Sequence['LinkTriple']] = None, clean: bool = True
    ) -> BackendNodeType:
        """Store the node in the database.

        :param links: optional links to add before storing
        :param clean: boolean, if True, will clean the attributes and extras before attempting to store
        """

    @abc.abstractmethod
    def clean_values(self):
        """Clean the values of the node fields.

        This method is called before storing the node.
        The purpose of this method is to convert data to a type which can be serialized and deserialized
        for storage in the DB without its value changing.
        """

    # attributes methods

    @property
    @abc.abstractmethod
    def attributes(self) -> Dict[str, Any]:
        """Return the complete attributes dictionary.

        .. warning:: While the entity is unstored, this will return references of the attributes on the database model,
            meaning that changes on the returned values (if they are mutable themselves, e.g. a list or dictionary) will
            automatically be reflected on the database model as well. As soon as the entity is stored, the returned
            attributes will be a deep copy and mutations of the database attributes will have to go through the
            appropriate set methods. Therefore, once stored, retrieving a deep copy can be a heavy operation. If you
            only need the keys or some values, use the iterators `attributes_keys` and `attributes_items`, or the
            getters `get_attribute` and `get_attribute_many` instead.

        :return: the attributes as a dictionary
        """

    @abc.abstractmethod
    def get_attribute(self, key: str) -> Any:
        """Return the value of an attribute.

        .. warning:: While the entity is unstored, this will return a reference of the attribute on the database model,
            meaning that changes on the returned value (if they are mutable themselves, e.g. a list or dictionary) will
            automatically be reflected on the database model as well. As soon as the entity is stored, the returned
            attribute will be a deep copy and mutations of the database attributes will have to go through the
            appropriate set methods.

        :param key: name of the attribute
        :return: the value of the attribute
        :raises AttributeError: if the attribute does not exist
        """

    def get_attribute_many(self, keys: Iterable[str]) -> List[Any]:
        """Return the values of multiple attributes.

        .. warning:: While the entity is unstored, this will return references of the attributes on the database model,
            meaning that changes on the returned values (if they are mutable themselves, e.g. a list or dictionary) will
            automatically be reflected on the database model as well. As soon as the entity is stored, the returned
            attributes will be a deep copy and mutations of the database attributes will have to go through the
            appropriate set methods. Therefore, once stored, retrieving a deep copy can be a heavy operation. If you
            only need the keys or some values, use the iterators `attributes_keys` and `attributes_items`, or the
            getters `get_attribute` and `get_attribute_many` instead.

        :param keys: a list of attribute names
        :return: a list of attribute values
        :raises AttributeError: if at least one attribute does not exist
        """
        try:
            return [self.get_attribute(key) for key in keys]
        except KeyError as exception:
            raise AttributeError(f'attribute `{exception}` does not exist') from exception

    @abc.abstractmethod
    def set_attribute(self, key: str, value: Any) -> None:
        """Set an attribute to the given value.

        :param key: name of the attribute
        :param value: value of the attribute
        """

    def set_attribute_many(self, attributes: Dict[str, Any]) -> None:
        """Set multiple attributes.

        .. note:: This will override any existing attributes that are present in the new dictionary.

        :param attributes: a dictionary with the attributes to set
        """
        for key, value in attributes.items():
            self.set_attribute(key, value)

    @abc.abstractmethod
    def reset_attributes(self, attributes: Dict[str, Any]) -> None:
        """Reset the attributes.

        .. note:: This will completely clear any existing attributes and replace them with the new dictionary.

        :param attributes: a dictionary with the attributes to set
        """

    @abc.abstractmethod
    def delete_attribute(self, key: str) -> None:
        """Delete an attribute.

        :param key: name of the attribute
        :raises AttributeError: if the attribute does not exist
        """

    def delete_attribute_many(self, keys: Iterable[str]) -> None:
        """Delete multiple attributes.

        :param keys: names of the attributes to delete
        :raises AttributeError: if at least one of the attribute does not exist
        """
        for key in keys:
            self.delete_attribute(key)

    @abc.abstractmethod
    def clear_attributes(self):
        """Delete all attributes."""

    @abc.abstractmethod
    def attributes_items(self) -> Iterable[Tuple[str, Any]]:
        """Return an iterator over the attributes.

        :return: an iterator with attribute key value pairs
        """

    @abc.abstractmethod
    def attributes_keys(self) -> Iterable[str]:
        """Return an iterator over the attribute keys.

        :return: an iterator with attribute keys
        """


class BackendNodeCollection(BackendCollection[BackendNode]):
    """The collection of `BackendNode` entries."""

    ENTITY_CLASS = BackendNode

    @abc.abstractmethod
    def get(self, pk: int):
        """Return a Node entry from the collection with the given id

        :param pk: id of the node
        """

    @abc.abstractmethod
    def delete(self, pk: int) -> None:
        """Remove a Node entry from the collection with the given id

        :param pk: id of the node to delete
        """



"""Data plugin that models a stashed folder on a remote computer."""

from typing import List, Tuple, Union

from aiida.common.datastructures import StashMode
from aiida.common.lang import type_check
from aiida.orm.fields import add_field

from .base import RemoteStashData

__all__ = ('RemoteStashFolderData',)


class RemoteStashFolderData(RemoteStashData):
    """Data plugin that models a folder with files of a completed calculation job that has been stashed through a copy.

    This data plugin can and should be used to stash files if and only if the stash mode is `StashMode.COPY`.
    """

    _storable = True

    __qb_fields__ = [
        add_field(
            'target_basepath',
            dtype=str,
            doc='The the target basepath',
        ),
        add_field(
            'source_list',
            dtype=List[str],
            doc='The list of source files that were stashed',
        ),
    ]

    def __init__(self, stash_mode: StashMode, target_basepath: str, source_list: List, **kwargs):
        """Construct a new instance

        :param stash_mode: the stashing mode with which the data was stashed on the remote.
        :param target_basepath: the target basepath.
        :param source_list: the list of source files.
        """
        super().__init__(stash_mode, **kwargs)
        self.target_basepath = target_basepath
        self.source_list = source_list

        if stash_mode != StashMode.COPY:
            raise ValueError('`RemoteStashFolderData` can only be used with `stash_mode == StashMode.COPY`.')

    @property
    def target_basepath(self) -> str:
        """Return the target basepath.

        :return: the target basepath.
        """
        return self.base.attributes.get('target_basepath')

    @target_basepath.setter
    def target_basepath(self, value: str):
        """Set the target basepath.

        :param value: the target basepath.
        """
        type_check(value, str)
        self.base.attributes.set('target_basepath', value)

    @property
    def source_list(self) -> Union[List, Tuple]:
        """Return the list of source files that were stashed.

        :return: the list of source files.
        """
        return self.base.attributes.get('source_list')

    @source_list.setter
    def source_list(self, value: Union[List, Tuple]):
        """Set the list of source files that were stashed.

        :param value: the list of source files.
        """
        type_check(value, (list, tuple))
        self.base.attributes.set('source_list', value)



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Work chain implementations for testing purposes."""

from aiida.common import AttributeDict
from aiida.engine import (
    BaseRestartWorkChain,
    ExitCode,
    ProcessHandlerReport,
    ToContext,
    WorkChain,
    append_,
    calcfunction,
    process_handler,
    while_,
    workfunction,
)
from aiida.engine.persistence import ObjectLoader
from aiida.orm import Int, List, Str
from aiida.plugins import CalculationFactory

ArithmeticAddCalculation = CalculationFactory('core.arithmetic.add')


class ArithmeticAddBaseWorkChain(BaseRestartWorkChain):
    """Ridiculous work chain around `AritmethicAddCalculation` with automated sanity checks and error handling."""

    _process_class = ArithmeticAddCalculation

    @classmethod
    def define(cls, spec):
        """Define the process specification."""
        super().define(spec)
        spec.expose_inputs(ArithmeticAddCalculation, namespace='add')
        spec.expose_outputs(ArithmeticAddCalculation)
        spec.outline(
            cls.setup,
            while_(cls.should_run_process)(
                cls.run_process,
                cls.inspect_process,
            ),
            cls.results,
        )
        spec.exit_code(100, 'ERROR_TOO_BIG', message='The sum was too big.')
        spec.exit_code(110, 'ERROR_ENABLED_DOOM', message='You should not have done that.')

    def setup(self):
        """Call the `setup` of the `BaseRestartWorkChain` and then create the inputs dictionary in `self.ctx.inputs`.

        This `self.ctx.inputs` dictionary will be used by the `BaseRestartWorkChain` to submit the process in the
        internal loop.
        """
        super().setup()
        self.ctx.inputs = AttributeDict(self.exposed_inputs(ArithmeticAddCalculation, 'add'))

    @process_handler(priority=500)
    def sanity_check_not_too_big(self, node):
        """My puny brain cannot deal with numbers that I cannot count on my hand."""
        if node.is_finished_ok and node.outputs.sum > 10:
            return ProcessHandlerReport(True, self.exit_codes.ERROR_TOO_BIG)

    @process_handler(priority=460, enabled=False)
    def disabled_handler(self, node):
        """By default this is not enabled and so should never be called, irrespective of exit codes of sub process."""
        return ProcessHandlerReport(True, self.exit_codes.ERROR_ENABLED_DOOM)

    @process_handler(priority=450, exit_codes=ExitCode(1000, 'Unicorn encountered'))
    def a_magic_unicorn_appeared(self, node):
        """As we all know unicorns do not exist so we should never have to deal with it."""
        raise RuntimeError('this handler should never even have been called')

    @process_handler(priority=400, exit_codes=ArithmeticAddCalculation.exit_codes.ERROR_NEGATIVE_NUMBER)
    def error_negative_sum(self, node):
        """What even is a negative number, how can I have minus three melons?!."""
        self.ctx.inputs.x = Int(abs(node.inputs.x.value))
        self.ctx.inputs.y = Int(abs(node.inputs.y.value))
        return ProcessHandlerReport(True)


class NestedWorkChain(WorkChain):
    """Nested workchain which creates a workflow where the nesting level is equal to its input."""

    @classmethod
    def define(cls, spec):
        super().define(spec)
        spec.input('inp', valid_type=Int)
        spec.outline(cls.do_submit, cls.finalize)
        spec.output('output', valid_type=Int, required=True)

    def do_submit(self):
        if self.should_submit():
            self.report('Submitting nested workchain.')
            return ToContext(workchain=append_(self.submit(NestedWorkChain, inp=self.inputs.inp - 1)))

    def should_submit(self):
        return int(self.inputs.inp) > 0

    def finalize(self):
        """Attach the outputs."""
        if self.should_submit():
            self.report('Getting sub-workchain output.')
            sub_workchain = self.ctx.workchain[0]
            self.out('output', Int(sub_workchain.outputs.output + 1).store())
        else:
            self.report('Bottom-level workchain reached.')
            self.out('output', Int(0).store())


class SerializeWorkChain(WorkChain):
    """Work chain that serializes inputs."""

    @classmethod
    def define(cls, spec):
        super().define(spec)

        spec.input('test', valid_type=Str, serializer=lambda x: Str(ObjectLoader().identify_object(x)))

        spec.outline(cls.echo)
        spec.outputs.dynamic = True

    def echo(self):
        self.out('output', self.inputs.test)


class NestedInputNamespace(WorkChain):
    """Work chain with nested namespace."""

    @classmethod
    def define(cls, spec):
        super().define(spec)

        spec.input('foo.bar.baz', valid_type=Int)
        spec.output('output', valid_type=Int)
        spec.outline(cls.do_echo)

    def do_echo(self):
        self.out('output', self.inputs.foo.bar.baz)


class ListEcho(WorkChain):
    """Work chain that simply echos a `List` input."""

    @classmethod
    def define(cls, spec):
        super().define(spec)

        spec.input('list', valid_type=List)
        spec.output('output', valid_type=List)

        spec.outline(cls.do_echo)

    def do_echo(self):
        self.out('output', self.inputs.list)


class DynamicNonDbInput(WorkChain):
    """Work chain with dynamic non_db inputs."""

    @classmethod
    def define(cls, spec):
        super().define(spec)
        spec.input_namespace('namespace', dynamic=True)
        spec.output('output', valid_type=List)
        spec.outline(cls.do_test)

    def do_test(self):
        input_list = self.inputs.namespace.input
        assert isinstance(input_list, list)
        assert not isinstance(input_list, List)
        self.out('output', List(list=list(input_list)).store())


class DynamicDbInput(WorkChain):
    """Work chain with dynamic input namespace."""

    @classmethod
    def define(cls, spec):
        super().define(spec)
        spec.input_namespace('namespace', dynamic=True)
        spec.output('output', valid_type=Int)
        spec.outline(cls.do_test)

    def do_test(self):
        input_value = self.inputs.namespace.input
        assert isinstance(input_value, Int)
        self.out('output', input_value)


class DynamicMixedInput(WorkChain):
    """Work chain with dynamic mixed input."""

    @classmethod
    def define(cls, spec):
        super().define(spec)
        spec.input_namespace('namespace', dynamic=True)
        spec.output('output', valid_type=Int)
        spec.outline(cls.do_test)

    def do_test(self):
        """Run the test."""
        input_non_db = self.inputs.namespace.inputs['input_non_db']
        input_db = self.inputs.namespace.inputs['input_db']
        assert isinstance(input_non_db, int)
        assert not isinstance(input_non_db, Int)
        assert isinstance(input_db, Int)
        self.out('output', Int(input_db + input_non_db).store())


class CalcFunctionRunnerWorkChain(WorkChain):
    """WorkChain which calls an InlineCalculation in its step."""

    @classmethod
    def define(cls, spec):
        super().define(spec)

        spec.input('input', valid_type=Int)
        spec.output('output', valid_type=Int)

        spec.outline(cls.do_run)

    def do_run(self):
        self.out('output', increment(self.inputs.input))


class WorkFunctionRunnerWorkChain(WorkChain):
    """WorkChain which calls a workfunction in its step"""

    @classmethod
    def define(cls, spec):
        super().define(spec)

        spec.input('input', valid_type=Str)
        spec.output('output', valid_type=Str)

        spec.outline(cls.do_run)

    def do_run(self):
        self.out('output', echo(self.inputs.input))


@workfunction
def echo(value):
    return value


@calcfunction
def increment(data):
    return Int(data + 1)



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Drop `db_dbcomputer.transport_params`

This is similar to migration django_0036

Revision ID: 07fac78e6209
Revises: de2eaf6978b4
Create Date: 2019-02-16 15:32:42.745450

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = '07fac78e6209'
down_revision = 'de2eaf6978b4'
branch_labels = None
depends_on = None


def upgrade():
    """Migrations for the upgrade."""
    op.drop_column('db_dbcomputer', 'transport_params')


def downgrade():
    """Migrations for the downgrade."""
    op.add_column(
        'db_dbcomputer',
        sa.Column('transport_params', postgresql.JSONB(astext_type=sa.Text()), autoincrement=False, nullable=True),
    )



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Module for the ProfileAccessManager that tracks process access to the profile."""

import contextlib
import os
import typing
from pathlib import Path

import psutil

from aiida.common.exceptions import LockedProfileError, LockingProfileError
from aiida.common.lang import type_check
from aiida.manage.configuration import Profile


class ProfileAccessManager:
    """Class to manage access to a profile.

    Any process that wants to request access for a given profile, should first call:

        ProfileAccessManager(profile).request_access()

    If this returns normally, the profile can be used safely. It will raise if it is locked, in which case the profile
    should not be used. If a process wants to request exclusive access to the profile, it should use ``lock``:

        with ProfileAccessManager(profile).lock():
            pass

    If the profile is already locked or is currently in use, an exception is raised.

    Access and locks of the profile will be recorded in a directory with files with a ``.pid`` and ``.lock`` extension,
    respectively. In principle then, at any one time, there can either be a number of pid files, or just a single lock
    file. If there is a mixture or there are more than one lock files, we are in an inconsistent state.
    """

    def __init__(self, profile: Profile):
        """Class that manages access and locks to the given profile.

        :param profile: the profile whose access to manage.
        """
        from aiida.manage.configuration.settings import ACCESS_CONTROL_DIR

        type_check(profile, Profile)
        self.profile = profile
        self.process = psutil.Process(os.getpid())
        self._dirpath_records = ACCESS_CONTROL_DIR / profile.name
        self._dirpath_records.mkdir(exist_ok=True)

    def request_access(self) -> None:
        """Request access to the profile.

        If the profile is locked, a ``LockedProfileError`` is raised. Otherwise a PID file is created for this process
        and the function returns ``None``. The PID file contains the command of the process.

        :raises ~aiida.common.exceptions.LockedProfileError: if the profile is locked.
        """
        error_message = (
            f'process {self.process.pid} cannot access profile `{self.profile.name}` ' f'because it is being locked.'
        )
        self._raise_if_locked(error_message)

        filepath_pid = self._dirpath_records / f'{self.process.pid}.pid'
        filepath_tmp = self._dirpath_records / f'{self.process.pid}.tmp'

        try:
            # Write the content to a temporary file and then move it into place with an atomic operation.
            # This prevents the situation where another process requests a lock while this file is being
            # written: if that was to happen, when the locking process is checking for outdated records
            # it will read the incomplete command, won't be able to correctly compare it with its running
            # process, and will conclude the record is old and clean it up.
            filepath_tmp.write_text(str(self.process.cmdline()))
            os.rename(filepath_tmp, filepath_pid)

            # Check again in case a lock was created in the time between the first check and creating the
            # access record file.
            error_message = (
                f'profile `{self.profile.name}` was locked while process ' f'{self.process.pid} was requesting access.'
            )
            self._raise_if_locked(error_message)

        except Exception as exc:
            filepath_tmp.unlink(missing_ok=True)
            filepath_pid.unlink(missing_ok=True)
            raise exc

    @contextlib.contextmanager
    def lock(self):
        """Request a lock on the profile for exclusive access.

        This context manager should be used if exclusive access to the profile is required. Access will be granted if
        the profile is currently not in use, nor locked by another process. During the context, the profile will be
        locked, which will be lifted automatically as soon as the context exits.

        :raises ~aiida.common.exceptions.LockingProfileError: if there are currently active processes using the profile.
        :raises ~aiida.common.exceptions.LockedProfileError: if there currently already is a lock on the profile.
        """
        error_message = (
            f'process {self.process.pid} cannot lock profile `{self.profile.name}` ' f'because it is already locked.'
        )
        self._raise_if_locked(error_message)

        self._clear_stale_pid_files()

        error_message = (
            f'process {self.process.pid} cannot lock profile `{self.profile.name}` ' f'because it is being accessed.'
        )
        self._raise_if_active(error_message)

        filepath = self._dirpath_records / f'{self.process.pid}.lock'
        filepath.touch()

        try:
            # Check if no other lock files were created in the meantime, which is possible if another
            # process was trying to obtain a lock at almost the same time.
            # By re-checking after creating the lock file we can ensure that racing conditions will never
            # cause two different processes to both think that they acquired the lock. It is still possible
            # that two processes that are trying to lock will think that the other acquired the lock first
            # and then both will fail, but this is a much safer case.
            error_message = (
                f'while process {self.process.pid} attempted to lock profile `{self.profile.name}`, '
                f'other process blocked it first.'
            )
            self._raise_if_locked(error_message)

            error_message = (
                f'while process {self.process.pid} attempted to lock profile `{self.profile.name}`, '
                f'other process started using it.'
            )
            self._raise_if_active(error_message)

            yield

        finally:
            filepath.unlink(missing_ok=True)

    def is_locked(self) -> bool:
        """Return whether the profile is locked."""
        return self._get_tracking_files('.lock', exclude_self=False) != []

    def is_active(self) -> bool:
        """Return whether the profile is currently in use."""
        return self._get_tracking_files('.pid', exclude_self=False) != []

    def clear_locks(self) -> None:
        """Clear all locks on this profile.

        .. warning:: This should only be used if the profile is currently still incorrectly locked because the lock was
            not automatically released after the ``lock`` contextmanager exited its scope.
        """
        for lock_file in self._get_tracking_files('.lock'):
            lock_file.unlink()

    def _clear_stale_pid_files(self) -> None:
        """Clear any stale PID files."""
        for path in self._get_tracking_files('.pid'):
            try:
                process = psutil.Process(int(path.stem))
            except psutil.NoSuchProcess:
                # The process no longer exists, so simply remove the PID file.
                path.unlink()
            else:
                # If the process exists but its command is different from what is written in the PID file,
                # we assume the latter is stale and remove it.
                if path.read_text() != str(process.cmdline()):
                    path.unlink()

    def _get_tracking_files(self, ext_string: str, exclude_self: bool = False) -> typing.List[Path]:
        """Return a list of all files that track the accessing and locking of the profile.

        :param ext_string:
            To get the files that track locking use `.lock`, to get the files that track access use `.pid`.

        :param exclude_self:
            If true removes from the returned list any tracking to the current process.
        """
        path_iterator = self._dirpath_records.glob('*' + ext_string)

        if exclude_self:
            filepath_self = self._dirpath_records / (str(self.process.pid) + ext_string)
            list_of_files = [filepath for filepath in path_iterator if filepath != filepath_self]

        else:
            list_of_files = list(path_iterator)

        return list_of_files

    def _raise_if_locked(self, message_start):
        """Raise a ``LockedProfileError`` if the profile is locked.

        :param message_start: Text to use as the start of the exception message.
        :raises ~aiida.common.exceptions.LockedProfileError: if the profile is locked.
        """
        list_of_files = self._get_tracking_files('.lock', exclude_self=True)

        if len(list_of_files) > 0:
            error_msg = message_start + '\nThe following processes are blocking the profile:\n'
            error_msg += '\n'.join(f' - pid {path.stem}' for path in list_of_files)
            raise LockedProfileError(error_msg)

    def _raise_if_active(self, message_start):
        """Raise a ``LockingProfileError`` if the profile is being accessed.

        :param message_start: Text to use as the start of the exception message.
        :raises ~aiida.common.exceptions.LockingProfileError: if the profile is active.
        """
        list_of_files = self._get_tracking_files('.pid', exclude_self=True)

        if len(list_of_files) > 0:
            error_msg = message_start + '\nThe following processes are accessing the profile:\n'
            error_msg += '\n'.join(f' - pid {path.stem} (`{path.read_text()}`)' for path in list_of_files)
            raise LockingProfileError(error_msg)



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Base SQLAlchemy models."""

from sqlalchemy import MetaData, event, inspect
from sqlalchemy.orm import declarative_base


class Model:
    """Base ORM model."""

    def __repr__(self) -> str:
        """Return a representation of the row columns"""
        string = f'<{self.__class__.__name__}'
        for col in self.__table__.columns:  # type: ignore[attr-defined]
            col_name = col.name
            if col_name == 'metadata':
                col_name = '_metadata'
            val = f'{getattr(self, col_name)!r}'
            if len(val) > 10:  # truncate long values
                val = val[:10] + '...'
            string += f' {col_name}={val},'
        return string + '>'


def instant_defaults_listener(target, args, kwargs):
    """Loop over the columns of the target model instance and populate defaults.

    SqlAlchemy does not set default values for table columns upon construction of a new instance,
    but will only do so when storing the instance.
    Any attributes that do not have a value but have a defined default,
    will be populated with this default.
    This does mean however, that before the instance is stored, these attributes are undefined, for example the UUID.

    This should be used as: https://docs.sqlalchemy.org/en/14/orm/events.html#sqlalchemy.orm.InstanceEvents.init
    """
    # using args would break this logic
    assert not args, f'args are not allowed in {target} instantiation'
    # If someone passes metadata in **kwargs we change it to _metadata
    if 'metadata' in kwargs:
        kwargs['_metadata'] = kwargs.pop('metadata')
    # don't allow certain JSON fields to be null
    for col in ('attributes', 'extras', '_metadata'):
        if col in kwargs and kwargs[col] is None:
            kwargs[col] = {}
    columns = inspect(target.__class__).columns
    # The only time that we allow mtime not to be non-null is when we explicitly pass it as a kwarg.
    # This covers the case that a node is constructed based on some very predefined data,
    # like when we create nodes in the AiiDA import functions
    if 'mtime' in columns and 'mtime' not in kwargs:
        kwargs['mtime'] = None
    for key, column in columns.items():
        if key in kwargs:
            continue
        if hasattr(column, 'default') and column.default is not None:
            if callable(column.default.arg):
                kwargs[key] = column.default.arg(target)
            else:
                kwargs[key] = column.default.arg


# see https://alembic.sqlalchemy.org/en/latest/naming.html
naming_convention = (
    ('pk', '%(table_name)s_pkey'),  # this is identical to the default PSQL convention
    ('ix', 'ix_%(table_name)s_%(column_0_N_label)s'),
    # note, indexes using varchar_pattern_ops should be named: 'ix_pat_%(table_name)s_%(column_0_N_label)s'
    ('uq', 'uq_%(table_name)s_%(column_0_N_name)s'),
    ('ck', 'ck_%(table_name)s_%(constraint_name)s'),
    ('fk', 'fk_%(table_name)s_%(column_0_N_name)s_%(referred_table_name)s'),
    # note, ideally we may also append with '_%(referred_column_0_N_name)s', but this causes ORM construction errors:
    # https://github.com/sqlalchemy/sqlalchemy/issues/5350
)

Base = declarative_base(cls=Model, name='Model', metadata=MetaData(naming_convention=dict(naming_convention)))
event.listen(Base, 'init', instant_defaults_listener, propagate=True)


def get_orm_metadata() -> MetaData:
    """Return the populated metadata object."""
    # we must load all models, to populate the ORM metadata

    return Base.metadata



###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Add a unique constraint on the UUID column of the Node model

Revision ID: 62fe0d36de90
Revises: 59edaf8a8b79
Create Date: 2018-07-02 17:50:42.929382

"""

from alembic import op

# revision identifiers, used by Alembic.
revision = '62fe0d36de90'
down_revision = '59edaf8a8b79'
branch_labels = None
depends_on = None


def upgrade():
    """Migrations for the upgrade."""
    from aiida.storage.psql_dos.migrations.utils.duplicate_uuids import verify_uuid_uniqueness

    verify_uuid_uniqueness('db_dbnode', op.get_bind())
    op.create_unique_constraint('db_dbnode_uuid_key', 'db_dbnode', ['uuid'])


def downgrade():
    """Migrations for the downgrade."""
    op.drop_constraint('db_dbnode_uuid_key', 'db_dbnode')



"""Module for file repository backend implementations."""

# AUTO-GENERATED

# fmt: off

from .abstract import *
from .disk_object_store import *
from .sandbox import *

__all__ = (
    'AbstractRepositoryBackend',
    'DiskObjectStoreRepositoryBackend',
    'SandboxRepositoryBackend',
)

# fmt: on
